/*
 * Code generation from interface descriptions.
 *
 * Copyright (C) 2019, 2020, 2022, 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <stdlib.h>
#include "client.h"
#include "common.h"
#include "config.h"
#include "declaration.h"
#include "dispatch.h"
#include "includes.h"
#include "interface.h"
#include "message.h"
#include "program.h"
#include "server.h"
#include "templates.h"



/* Client, server and common files. */

FILE *client_fp = NULL, *client_header_fp = NULL,
     *server_fp = NULL, *server_header_fp = NULL,
     *interface_fp = NULL;



/* Generate functions corresponding to the signatures. */

static void write_functions(struct signature *sig, FILE *fp,
                            struct interface *iface,
                            void (*write_function)(struct signature *, FILE *, struct interface *))
{
  if (sig == NULL)
    return;

  write_function(sig, fp, iface);
  write_functions(sig->tail, fp, iface, write_function);
}

/* Write includes for base interfaces. */

static void write_server_includes(struct interface *iface, FILE *fp)
{
  struct interface_ref *base;

  for (base = iface->bases; base != NULL; base = base->tail)
    write_include(base->iface->output_basename, "_server.h", fp);
}



/* Generate files containing interface details. */

void write_files(struct interface *iface)
{
  char *filename;

  if (conf.headers)
  {
    if (conf.client)
    {
      client_header_fp = get_output_file(client_header_filename,
                                         iface->output_dirname,
                                         iface->output_basename);
      if (client_header_fp == NULL)
        goto finalisation;
    }

    if (conf.server)
    {
      server_header_fp = get_output_file(server_header_filename,
                                         iface->output_dirname,
                                         iface->output_basename);
      if (server_header_fp == NULL)
        goto finalisation;
    }
  }

  if (conf.interfaces)
  {
    interface_fp = get_output_file(interface_filename, iface->output_dirname,
                                   iface->output_basename);
    if (interface_fp == NULL)
      goto finalisation;
  }

  if (conf.routines)
  {
    if (conf.client)
    {
      filename = (conf.language == CPP_LANGUAGE) ? client_filename_cpp
                                                 : client_filename_c;
      client_fp = get_output_file(filename, iface->output_dirname,
                                  iface->output_basename);
      if (client_fp == NULL)
        goto finalisation;
    }

    if (conf.server)
    {
      filename = (conf.language == CPP_LANGUAGE) ? server_filename_cpp
                                                 : server_filename_c;
      server_fp = get_output_file(filename, iface->output_dirname,
                                  iface->output_basename);
      if (server_fp == NULL)
        goto finalisation;
    }
  }

  /* Emit prologues. */

  if (client_fp != NULL)
    fprintf(client_fp, client_prologue, iface->output_basename,
            iface->output_basename);

  if (client_header_fp != NULL)
    fprintf(client_header_fp, header_prologue, "CLIENT", iface->name,
                                               "CLIENT", iface->name);

  if (server_fp != NULL)
    fprintf(server_fp, server_prologue, iface->output_basename,
            iface->output_basename);

  if (server_header_fp != NULL)
    fprintf(server_header_fp, server_header_prologue, iface->name, iface->name);

  if (interface_fp != NULL)
    fprintf(interface_fp, header_prologue, "INTERFACE", iface->name,
                                           "INTERFACE", iface->name);

  /* Write the interface details. */

  write_interfaces(iface);

finalisation:

  /* Close the files. */

  if (client_fp != NULL)
    fclose(client_fp);

  if (client_header_fp != NULL)
  {
    fprintf(client_header_fp, header_epilogue);
    fclose(client_header_fp);
  }

  if (server_fp != NULL)
    fclose(server_fp);

  if (server_header_fp != NULL)
  {
    fprintf(server_header_fp, header_epilogue);
    fclose(server_header_fp);
  }

  if (interface_fp != NULL)
  {
    fprintf(interface_fp, header_epilogue);
    fclose(interface_fp);
  }
}

/* Generate source file content corresponding to each interface in turn. */

void write_interfaces(struct interface *iface)
{
  if (iface == NULL)
    return;

  /* The list is reversed by recursing to subsequent interfaces and thus
     processing them first. */

  write_interfaces(iface->tail);

  /* Emit function definitions. */

  if (client_fp != NULL)
  {
    /* Do not write client definitions for compound interfaces. */

    if (!is_compound_interface(iface))
    {
      write_functions(iface->signatures, client_fp, iface,
                      write_client_function);

      if (conf.language == C_LANGUAGE)
        write_client_interface(iface, client_fp);
    }
  }

  if (server_fp != NULL)
  {
    /* Write the dispatcher and handler functions. */

    write_functions(iface->signatures, server_fp, iface,
                    write_server_function);

    /* Emit handle and dispatch functions. */

    write_handler_signature(iface->name, DEFINITION_ROLE, server_fp);
    fprintf(server_fp, handle_function, iface->name);
    fputs(END_FUNCTION, server_fp);

    write_dispatcher(server_fp, iface);

    /* Emit default configuration instance definition. */

    fprintf(server_fp, server_config_instance,
            iface->name, iface->name, iface->name);
  }

  /* Emit the headers. */

  if (client_header_fp != NULL)
  {
    /* Emit interface include to obtain the type employed by the client
       interface, doing this once per file. */

    if (iface->tail == NULL)
      write_include(iface->output_basename, "_interface.h", client_header_fp);

    /* Do not write client definitions for compound interfaces. */

    if (!is_compound_interface(iface))
      write_interface_definition(iface, CLIENT_ROLE, client_header_fp);
  }

  if (server_header_fp != NULL)
  {
    /* Emit interface include to obtain the interface type, doing this once per
       file. */

    if (iface->tail == NULL)
      write_include(iface->output_basename, "_interface.h", server_header_fp);

    /* Write any includes for base interfaces. */

    write_server_includes(iface, server_header_fp);

    /* Emit signatures. */

    write_signatures(iface->signatures, DECLARATION_ROLE, server_header_fp,
                     iface, write_server_signature);

    /* Emit dispatch and handle function signatures. */

    write_dispatcher_signature(iface->name, DECLARATION_ROLE, server_header_fp);
    write_handler_signature(iface->name, DECLARATION_ROLE, server_header_fp);

    /* Emit default configuration instance declaration. */

    fprintf(server_header_fp, server_config_instance_declaration, iface->name);
  }

  /* The server interface is used by client and server code. */

  if (interface_fp != NULL)
    write_interface_definition(iface, SERVER_ROLE, interface_fp);
}

/* vim: tabstop=2 expandtab shiftwidth=2
*/
